# -*- coding: utf8 -*-

from qt import *
from qspinwidget import QSpinWidget
from rangecontrol import RangeControl
from types import IntType


class SpinBoxIntValidator(QIntValidator):
    def __init__(self, sb, name):
        QIntValidator.__init__(self, sb, name)
        self.spinBox = sb

    def validate(self, text, pos):
        pref = self.spinBox.prefix()
        suff = self.spinBox.suffix()
        suffStriped = suff.stripWhiteSpace()
        overhead = pref.length() + suff.length()
        state = QValidator.Invalid
        self.setRange(self.spinBox.minValue(),
                      self.spinBox.maxValue())
        if overhead == 0:
            state, pos = QIntValidator.validate(self, text, pos)
        else:
            if text.length() >= overhead and text.startsWith(pref) and \
                 (text.endsWith(suff) or text.endsWith(suffStriped)):
                if text.endsWith(suffStriped):
                    overhead = pref.length() + suffStriped.length()
                core = text.mid(pref.length(), text.length() - overhead)
                corePos = pos - pref.length()
                state, corePos = QIntValidator.validate(self, core, corePos)
                pos = corePos + pref.length()
                text.replace(pref.length(), text.length() - overhead, core)
            else:
                state, pos = QIntValidator.validate(self, text, pos)
                if state == QValidator.Invalid:
                    special = self.spinBox.specialValueText().stripWhiteSpace()
                    candidate = text.stripWhiteSpace()
                    if special.startsWith(candidate):
                        if candidate.length() == special.length():
                            state = QValidator.Acceptable
                        else:
                            state = QValidator.Intermediate
        return state, pos


class SpinBox(QWidget, RangeControl):
    def __init__(self, *args):
        # minv, maxv, step, page, val
        defargs = [0, 99, 1, 10, 0]
        parent, name = None, ""
        # convert args tuple to list
        args = map(None, args)
        # proces up to 5 int args
        for i in range(5):
            # stop, if None or QObject derived argument appears
            if args and args[0] != None:
                try:
                    args[0].inherits("QObject")
                    break
                except AttributeError:
                    v = args.pop(0)
                    if type(v) != IntType:
                        raise "SpinBox: int type for argument %s expected: got %s (%s)" % (i + 1, v, type(v))
                    defargs[i] = v
            else:
                break
        if args:
            parent = args.pop(0)
        if args:
            name = args.pop(0)
        # may check for superfluous arguments?

        QWidget.__init__(self, parent, name, Qt.WNoAutoErase)
        RangeControl.__init__(self, *defargs)

        self.controls = QSpinWidget(self, "spinbox_controls")
        self.connect(self.controls, SIGNAL("stepUpPressed()"), self.stepUp)
        self.connect(self.controls, SIGNAL("stepDownPressed()"), self.stepDown)
        self.wrap = False
        self.edited = False
        self.selreq = False
        self.pfix = QString()
        self.sfix = QString()
        self.specText = QString()
        self.validate = SpinBoxIntValidator(self, "validator")
        self.vi = QLineEdit(self, "spinbox_edit")
        self.controls.setEditWidget(self.vi)
        self.vi.setValidator(self.validate)
        self.vi.installEventFilter(self)
        self.vi.setFrame(False)
        self.setFocusProxy(self.vi)
        self.setSizePolicy(QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed))
        self.setBackgroundMode(Qt.PaletteBackground, Qt.PaletteBase)
        self.updateDisplay()
        self.connect(self.vi, SIGNAL("textChanged(const QString&)"), self.textChanged)

    def text(self):
        return self.vi.text()

    def cleanText(self):
        s = self.vi.text().stripWhiteSpace()
        if self.pfix:
            px = self.pfix.stripWhiteSpace()
            len = px.length()
            # Remove _only_ if it is the prefix
            if len and s.left(len) == px:
                s.remove(0, len)
        if self.sfix:
            sx = self.sfix.stripWhiteSpace()
            len = sx.length()
            # Remove _only_ if it is the suffix
            if len and s.right(len) == sx:
                s.truncate(s.length() - len)
        return s.stripWhiteSpace()

    def setSpecialValueText(self, text):
        self.specText = QString(text)
        self.updateDisplay()

    def specialValueText(self):
        return QString(self.specText)

    def setPrefix(self, text):
        self.pfix = QString(text)
        self.updateDisplay()

    def prefix(self):
        return QString(self.pfix)

    def setSuffix(self, text):
        self.sfix = QString(text)
        self.updateDisplay()

    def suffix(self):
        return QString(self.sfix)

    def setWrapping(self, b):
        self.wrap = bool(b)
        self.updateDisplay()

    def wrapping(self):
        return self.wrap

    def stepUp(self):
        if self.edited:
            self.interpretText()
        if self.wrap and self.value() + self.lineStep() > self.maxValue():
            self.setValue(self.minValue())
        else:
            self.addLine()

    def stepDown(self):
        if self.edited:
            self.interpretText()
        if self.wrap and self.value() - self.lineStep() < self.minValue():
            self.setValue(self.maxValue())
        else:
            self.subtractLine()

    def pageUp(self):
        if self.edited:
            self.interpretText()
        if self.wrap and self.value() + self.pageStep() > self.maxValue():
            self.setValue(self.minValue())
        else:
            self.addPage()

    def pageDown(self):
        if self.edited:
            self.interpretText()
        if self.wrap and self.value() - self.pageStep() < self.minValue():
            self.setValue(self.maxValue())
        else:
            self.subtractPage()

    def setLineStep(self, i):
        self.setSteps(i, self.pageStep())

    def setPageStep(self, i):
        self.setSteps(self.lineStep(), i)

    def setValidator(self, v):
        if self.vi:
            self.vi.setValidator(v)

    def validator(self):
        return self.vi and self.vi.validator() or None

    def editor(self):
        return self.vi

    def setValue(self, value):
        # we ignore anything entered and not yet interpreted
        self.edited = False
        RangeControl.setValue(self, value)

    def value(self):
        if self.edited:
            self.interpretText()
        return RangeControl.value(self)

    # slot called from self.vi.textChanged
    def textChanged(self):
        self.edited = True

    # triggered by RangeControl
    def valueChange(self):
        self.selreq = self.hasFocus()
        self.updateDisplay()
        self.selreq = False
        self.emit(PYSIGNAL("valueChanged"), (self.value(),))
        self.emit(PYSIGNAL("textChanged"), (self.currentValueText(),))

    # triggered by RangeControl
    def rangeChange(self):
        self.updateDisplay()

    def selectAll(self):
        overhead = self.pfix.length() + self.sfix.length()
        if not overhead or self.currentValueText() == self.specText:
            self.vi.selectAll()
        else:
            self.vi.setSelection(self.pfix.length(), self.vi.text().length() - overhead)

    def mapValueToText(self, v):
        return QString().setNum(v)

    def mapTextToValue(self):
        newVal, ok = self.text().toInt()
        if not ok and (self.pfix or self.sfix):
            # Try removing any pre/suffix
            newVal, ok = self.cleanText().toInt()
        #print "SpinBox.mapTextToValue:", newVal, ok
        return newVal, ok

    def currentValueText(self):
        if self.value() == self.minValue() and self.specText:
            s = QString(self.specText)
        else:
            s = QString(self.pfix) + self.mapValueToText(self.value()) + self.sfix
        return s

    def interpretText(self):
        ok = True
        done = False
        newVal = 0
        if self.specText:
            s = self.text().stripWhiteSpace()
            t = self.specText.stripWhiteSpace()
            if s == t:
                newVal = self.minValue()
                done = True
        if not done:
            newVal, ok = self.mapTextToValue()
        if ok:
            self.setValue(newVal)
            self.updateDisplay()

    def updateDisplay(self):
        self.vi.setUpdatesEnabled(False)
        self.vi.setText(self.currentValueText())
        if self.selreq and self.isVisible() and (self.hasFocus() or self.vi.hasFocus()):
            self.selectAll()
        else:
            if self.sfix and self.vi.text().endsWith(self.sfix):
                self.vi.setCursorPosition(self.vi.text().length() - self.sfix.length())
        self.vi.setUpdatesEnabled(True)
        self.vi.repaint(False)  # immediate repaint needed for some reason
        self.edited = False

        upEnabled = self.isEnabled() and (self.wrap or self.value() < self.maxValue())
        downEnabled = self.isEnabled() and (self.wrap or self.value() > self.minValue())
        self.controls.setUpEnabled(upEnabled)
        self.controls.setDownEnabled(downEnabled)
        self.vi.setEnabled(self.isEnabled())
        self.repaint(False)

    def sizeHint(self):
        self.constPolish()
        sz = self.vi.sizeHint()
        h = sz.height()
        fm = QFontMetrics(self.font())
        w = 35
        wx = fm.width(" ") * 2
        s = self.pfix + self.mapValueToText(self.minValue()) + self.sfix
        w = max(w, fm.width(s) + wx)
        s = self.pfix + self.mapValueToText(self.maxValue()) + self.sfix
        w = max(w, fm.width(s) + wx)
        if self.specText:
            w = max(w, fm.width(self.specText) + wx)
        sz = self.style().sizeFromContents(QStyle.CT_SpinBox, self,
                        QSize(w + self.controls.downRect().width(),
                              h + self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
                                * 2).expandedTo(QApplication.globalStrut()))
        return sz

    def minimumSizeHint(self):
        w = self.vi.minimumSizeHint().width() + self.controls.downRect().width()
        h = max(self.vi.minimumSizeHint().height(), self.controls.minimumSizeHint().height())
        sz = QSize(w, h)
        return sz

    def eventFilter(self, obj, ev):
        if obj != self.vi:
            return QWidget.eventFilter(self, obj, ev)

        if ev.type() == QEvent.KeyPress:
            key = ev.key()
            if key == Qt.Key_Tab or key == Qt.Key_BackTab:
                if ev.state() & Qt.ControlButton:
                    return False
                if self.edited:
                    self.interpretText()
                qApp.sendEvent(self, ev)
                return True
            elif key == Qt.Key_Up:
                self.stepUp()
                return True
            elif key == Qt.Key_Down:
                self.stepDown()
                return True
            elif key == Qt.Key_PageUp:
                self.pageUp()
                return True
            elif key == Qt.Key_PageDown:
                self.pageDown()
                return True
            elif key == Qt.Key_Enter or key == Qt.Key_Return:
                self.interpretText()
                return False
        elif ev.type() == QEvent.FocusOut or ev.type() == QEvent.Hide:
            if self.edited:
                self.interpretText()
        return False

    def resizeEvent(self, ev):
        self.controls.resize(self.width(), self.height())

    def wheelEvent(self, e):
        e.accept()
        self.offset = 0.0
        self.offset_owner = None
        if self.offset_owner != self:
            self.offset_owner = self
            self.offset = 0

        self.offset += -e.delta()/120
        if abs(self.offset) < 1:
            return
        ioff = int(self.offset)
        for i in range(abs(ioff)):
            self.offset > 0 and self.stepDown() or self.stepUp()
            self.offset -= ioff

    def setEnabled(self, enabled):
        QWidget.setEnabled(self, enabled)
        self.updateDisplay()

    def styleChange(self, old):
        self.arrangeWidgets()
        QWidget.styleChange(self, old)

    # QSpinWidget mgmt
    def arrangeWidgets(self):
        self.controls.arrange()

    def upRect(self):
        return self.controls.upRect()

    def downRect(self):
        return self.controls.downRect()

    def setButtonSymbols(self, newSymbols):
        if self.buttonSymbols() == newSymbols:
            return
        if newSymbols in (QSpinWidget.UpDownArrows, QSpinWidget.PlusMinus):
            self.controls.setButtonSymbols(newSymbols)

    def buttonSymbols(self):
        bsyms = d.controls.buttonSymbols()
        if bsyms in (QSpinWidget.UpDownArrows, QSpinWidget.PlusMinus):
            return bsyms
        return QSpinWidget.UpDownArrows


class SpinBoxDecimalValidator(QDoubleValidator):
    def __init__(self, sb, name):
        QDoubleValidator.__init__(self, sb, name)
        self.spinBox = sb

    def validate(self, text, pos):
        pref = self.spinBox.prefix()
        suff = self.spinBox.suffix()
        suffStriped = suff.stripWhiteSpace()
        overhead = pref.length() + suff.length()
        state = QValidator.Invalid
        self.setRange(float(self.spinBox.minValue()),
                      float(self.spinBox.maxValue()),
                      self.spinBox.getDecimals())
        if overhead == 0:
            state, pos = QDoubleValidator.validate(self, self.spinBox.unformatValue(text), pos)
        else:
            if text.length() >= overhead and text.startsWith(pref) and \
                 (text.endsWith(suff) or text.endsWith(suffStriped)):
                if text.endsWith(suffStriped):
                    overhead = pref.length() + suffStriped.length()
                core = text.mid(pref.length(), text.length() - overhead)
                corePos = pos - pref.length()
                state, corePos = QDoubleValidator.validate(self, self.spinBox.unformatValue(core), corePos)
                pos = corePos + pref.length()
                text.replace(pref.length(), text.length() - overhead, core)
            else:
                state, pos = QDoubleValidator.validate(self, self.spinBox.unformatValue(text), pos)
                if state == QValidator.Invalid:
                    special = self.spinBox.specialValueText().stripWhiteSpace()
                    candidate = text.stripWhiteSpace()
                    if special.startsWith(candidate):
                        if candidate.length() == special.length():
                            state = QValidator.Acceptable
                        else:
                            state = QValidator.Intermediate
        return state, pos


class DecimalSpinBox(SpinBox):
    def __init__(self, *args, **kwargs):
        # minv, maxv, step, page, val
        defargs = [0, 99, Decimal("0.1"), Decimal("10.0"), Decimal("0.00")]
        parent, name = None, ""
        # convert args tuple to list
        args = map(None, args)
        # proces up to 5 int/decimal args
        for i in range(5):
            # stop, if None or QObject derived argument appears
            if args and args[0] != None:
                try:
                    args[0].inherits("QObject")
                    break
                except AttributeError:
                    defargs[i] = Decimal(args.pop(0))
            else:
                break
        if args:
            parent = args.pop(0)
        if args:
            name = args.pop(0)
        # may check for superfluous arguments?

        self.decimals = 0
        self.decimalpoint = '.'
        self.separator = ','
        for key in kwargs.keys():
            if key in ('decimals', 'decimalpoint', 'separator'):
                setattr(self, key, kwargs[key])

        SpinBox.__init__(self, parent, name)
        RangeControl.__init__(self, *defargs)
        if self.decimals == 0:
            d = self.value().as_tuple()[2]
            if d < 0:
                self.decimals = abs(d)
        self.setValidator(SpinBoxDecimalValidator(self, "validator"))

    def getDecimals(self):
        return self.decimals

    def setDecimals(self, decimals):
        self.decimals = int(decimals)

    def getDecimalPoint(self):
        return self.decimalpoint
    
    def setDecimalPoint(self, decimalpoint):
        self.decimalpoint = decimalpoint

    def getSeparator(self):
        return self.separator
        
    def setSeparator(self, separator):
        self.separator = separator

    def mapValueToText(self, value):
        return self.formatValue(self.convertValue(value))

    def mapTextToValue(self):
        return self.convertValue(self.unformatValue(self.cleanText())), True

    def convertValue(self, value):
        return Decimal(value).quantize(Decimal((0, (1,), -self.decimals)))

    def unformatValue(self, text):
        text = unicode(text)
        if self.separator:
            text = text.replace(self.separator, "")
        if self.decimalpoint != ".":
            text = text.replace(self.decimalpoint, ".")
        return text.strip()

    def formatValue(self, value):
        q = Decimal((0, (1,), -self.decimals))
        sign, digits, exp = value.quantize(q).as_tuple()
        assert exp == -self.decimals
        result = []
        digits = map(str, digits)
        build, next = result.append, digits.pop
        for i in range(self.decimals):
            if digits:
                build(next())
            else:
                build('0')
        if result:
            build(self.decimalpoint)
        if len(digits):
            i = 0
            while digits:
                build(next())
                i += 1
                if i == 3 and digits:
                    i = 0
                    build(self.separator)
        else:
            build('0')
        if sign:
            build('-')
        result.reverse()
        return ''.join(result)


if __name__ == '__main__':
    import sys, codecs
    sys.stdout = codecs.lookup('utf8')[-1](sys.stdout, "ignore")
    sys.stderr = codecs.lookup('utf8')[-1](sys.stderr, "ignore")

    from decimal import Decimal
    
    class testSpinBox(QWidget):
        def __init__(self, *args):
            QWidget.__init__(self, *args)
            layout = QGridLayout(self, 1, 1, 11, 6, "layout")
            deflimits = (-5, 155, 1, 10, 0)
            limits = { 
                2: (-5, 5555, Decimal("0.1"), 10, Decimal("-5.00")),
                3: (-1000000000000, 1000000000000, 100000, 1000000000, Decimal("0.000000000")),
                4: (-5, 1550000, Decimal("0.0001"), 1000, Decimal("-5.0000")),
                5: (-9999999, 9999999, Decimal("0.1"), 1, Decimal("7.9")),
                6: (-9999999, 9999999, Decimal("0.001"), 10, Decimal("17.923")),
                7: (-9999999, 9999999, Decimal("0.01"), 100, Decimal("27.89")),
                9: (-10000, 10000, Decimal("0.0001"), 100, Decimal("0.0000")),
            }
            prefixes = { 5: "from ", 6: "to ", 9: "prefix: " }
            suffixes = { 1: " items", 5: " km", 6: " km", 7: u" €", 8: " mm", 9: " suffix" }
            extra = { "decimalpoint": ",", "separator": "." }
            for i in range(1, 11):
                l = QLabel("SpinBox %s:" % i, self)
                layout.addWidget(l, i, 0)
                minv, maxv, line, page, val = limits.get(i, deflimits)
                if type(val) == IntType:
                    sb = SpinBox(minv, maxv, line, page, val, self, "sb%s" % i)
                else:
                    sb = DecimalSpinBox(minv, maxv, line, page, val, self, "sb%s" % i, **extra)
                sb.setPrefix(prefixes.get(i, ""))
                sb.setSuffix(suffixes.get(i, ""))
                if i == 2:
                    sb.setSpecialValueText("lowest value")
                layout.addWidget(sb, i, 1)
                self.connect(sb, PYSIGNAL("valueChanged"), self.valueChanged)

        def valueChanged(self, value):
            print u"value changed[%s]: %s <%s>" % (self.sender().name(), value, self.sender().currentValueText())

    app = QApplication([])
    t = testSpinBox()
    t.resize(t.sizeHint())
    t.show()
    app.setMainWidget(t)
    app.exec_loop()
